(*
  Copyright 2010 Gerd Stolpmann

  This file is part of Plasma, a distributed filesystem and a
  map/reduce computation framework. Unless you have a written license
  agreement with the copyright holder (Gerd Stolpmann), the following
  terms apply:

  Plasma is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Plasma is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.

*)
(* $Id: mapred_main.ml 282 2010-10-28 00:13:50Z gerd $ *)

open Printf

let this =
  Filename.basename Sys.argv.(0)

let usage() =
  eprintf "usage: %s <mode> [-help|options]\n" this;
  prerr_endline "";
  prerr_endline "  available modes:";
  prerr_endline "  - start_task_servers";
  prerr_endline "  - stop_task_servers";
  prerr_endline "  - exec_job"


let start_task_servers mj =
  let cf_name = ref (Sys.argv.(0) ^ ".conf") in
  Arg.parse
    [ "-conf", Arg.Set_string cf_name,
      "<file>   Set the conf file ";
    ]
    (fun s -> raise(Arg.Bad ("Extra argument: "  ^ s)))
    (sprintf "usage: %s start_task_servers [options]" this);

  let cf = Netplex_config.read_config_file !cf_name in
  let conf = Mapred_config.extract_config cf in
  
  (* Generate config file of the task server: *)
  let ttmp = conf # mr_task_tmpdir in
  let n, f = Filename.open_temp_file "plasmamr_task" ".conf" in
  try
    fprintf f "netplex {\n";
    fprintf f "  controller {\n";
    fprintf f "    socket_directory = \"%s/sockdir\";\n" ttmp;
    fprintf f "    max_level = \"debug\";\n";
    fprintf f "    logging {\n";
    fprintf f "       type = \"multi_file\";\n";
    fprintf f "       directory = \"%s\";\n" ttmp;
    fprintf f "       file { file = \"task_server_debug.log\"; max_level = \"debug\" };\n";
    fprintf f "       file { file = \"task_server_info.log\"; max_level = \"info\" };\n";
    fprintf f "       file { file = \"task_server_crit.log\"; max_level = \"crit\" }\n";
    fprintf f "    }\n";
    fprintf f "  };\n";
    fprintf f "  service {\n";
    fprintf f "    name = \"Mapred_task_server\";\n";
    fprintf f "    protocol {\n";
    fprintf f "      name = \"RPC\";\n";
    fprintf f "      address {\n";
    fprintf f "        type = \"internet\";\n";
    fprintf f "        bind = \"0.0.0.0:%d\"\n" conf#mr_task_port;
    fprintf f "      }\n";
    fprintf f "    };\n";
    fprintf f "    processor { \n";
    fprintf f "      type = \"mapred_task_server\"\n";
    fprintf f "    };\n";
    fprintf f "    workload_manager {\n";
    fprintf f "      type = \"dynamic\";\n";
    fprintf f "      max_jobs_per_thread = 1;\n";
    fprintf f "      min_free_jobs_capacity = 1;\n";
    fprintf f "      max_free_jobs_capacity = 1;\n";
    fprintf f "      max_threads = 100\n";
    fprintf f "    }\n";
    fprintf f "  }\n";
    fprintf f "}\n";
    close_out f;
  
    let files_to_copy =
      [ n,            "task_server.conf";
	Sys.argv.(0), "task_server";
	!cf_name,     "mapred.conf"
      ] in
    List.iter
      (fun hostname ->
	 List.iter
	   (fun (localname, remotename) ->
	      let cmd = 
		sprintf "scp %s %s:%s/%s" 
		  localname hostname ttmp remotename in
	      let code = Sys.command cmd in
	      if code <> 0 then
		failwith ("Exec failed: " ^ cmd);
	   )
	   files_to_copy;
      )
      conf#mr_task_nodes;
    Sys.remove n;

    List.iter
      (fun hostname ->
	 let cmd =
	   sprintf "ssh %s %s/task_server task_server \
                   -conf %s/task_server.conf -mapred-conf %s/mapred.conf \
                   -pid %s/task_server.pid '2>%s/task_server.stderr' '>&2'"
	     hostname
	     ttmp ttmp ttmp ttmp ttmp in
	 let code = Sys.command cmd in
	 if code <> 0 then
	   failwith ("Exec failed: " ^ cmd);
      )
      conf#mr_task_nodes;
    print_endline "Servers started"
  with
    | error ->
	close_out f;
	Sys.remove n;
	raise error


let stop_task_servers mj =
  let cf_file = ref (Sys.argv.(0) ^ ".conf") in
  Arg.parse
    [ "-conf", Arg.Set_string cf_file,
      "<file>   Set the conf file ";
    ]
    (fun s -> raise(Arg.Bad ("Extra argument: "  ^ s)))
    (sprintf "usage: %s stop_task_servers [options]" this);

  let cf = Netplex_config.read_config_file !cf_file in
  let conf = Mapred_config.extract_config cf in
  let ttmp = conf # mr_task_tmpdir in
  
  List.iter
    (fun hostname ->
       let cmd =
	 sprintf "ssh %s 'kill $(cat %s/task_server.pid)'"
	   hostname 
	   ttmp in
       let code = Sys.command cmd in
       if code <> 0 then
	 prerr_endline ("Exec failed: " ^ cmd);
    )
    conf#mr_task_nodes;
  print_endline "Servers killed"

       
let task_server mj =
  let netplex_args, netplex_cmdconf = Netplex_main.args () in
  let mapred_conf = ref "mapred.conf" in
  Arg.parse
    ( [ "-mapred-conf", Arg.Set_string mapred_conf,
	"<file>   Set the mapred conf file";
      ] @
	netplex_args
    )
    (fun s -> raise(Arg.Bad ("Extra argument: "  ^ s)))
    (sprintf "usage: %s task_server [options]" this);

  let mapred_cf = Netplex_config.read_config_file !mapred_conf in

  let parallelizer =
    Netplex_mp.mp() in
  
  prerr_endline "Startup";
  Netlog.logf `Notice "Startup";

  Netplex_main.startup
    parallelizer
    Netplex_log.logger_factories
    Netplex_workload.workload_manager_factories
    [ Mapred_task_server.factory mj mapred_cf ]
    netplex_cmdconf


let exec_job mj =
  let cf_file = ref (Sys.argv.(0) ^ ".conf") in
  let dump_plan = ref false in
  let keep_temp_files = ref false in
  let args = ref [] in

  let proc_override s =
    try
      match Pcre.extract ~pat:"^([a-zA-Z0-9_]+)=(.*)$" ~full_match:false s with
	| [| n; v |] -> args := !args @ [ n,v ]
	| _ -> raise Not_found
    with
      | Not_found ->
	  raise (Arg.Bad ("Cannot parse this override: " ^ s)) in

  Arg.parse
    [ "-conf", Arg.Set_string cf_file,
      "<file>   Set the conf file ";

      "-override", Arg.String proc_override,
      "<name>=<value>    Override a parameter from the 'mapredjob' section";

      "-keep-temp-files", Arg.Set keep_temp_files,
      "    do not delete intermediate files";

      "-dump-plan", Arg.Set dump_plan,
      "    Dump the plan before mapping and after mapping";

      "-debug", Arg.String (fun s -> Netlog.Debug.enable_module s),
      "<module>  Enable debug messages for <module>";

      "-debug-all", Arg.Unit (fun () -> Netlog.Debug.enable_all()),
      "  Enable all debug messages";

      "-debug-list", Arg.Unit (fun () -> 
                                 List.iter print_endline (Netlog.Debug.names());
                                 exit 0),
      "  Show possible modules for -debug, then exit"

    ]
    (fun s -> raise(Arg.Bad ("Extra argument: "  ^ s)))
    (sprintf "usage: %s stop_task_servers [options]" this);

  let cf = Netplex_config.read_config_file !cf_file in
  let conf = Mapred_config.extract_config cf in
  let (jc,mjc) = 
    Mapred_job_config.extract_job_config cf !args mj#custom_params in

  printf "--------------------------------------------------------\n";
  printf "Job name:            %s\n" jc#name;
  printf "Job ID:              %s\n" jc#job_id;
  printf "PlasmaFS input dir:  %s\n" jc#input_dir;
  printf "PlasmaFS output dir: %s\n" jc#output_dir;
  printf "PlasmaFS work dir:   %s\n" jc#work_dir;
  printf "PlasmaFS log dir:    %s\n" jc#log_dir;
  printf "Number task nodes:   %d\n" (List.length conf#mr_task_nodes);
  printf "Task server dir:     %s\n" conf#mr_task_tmpdir;
  printf "--------------------------------------------------------\n%!";

  let open_cluster esys =
    Plasma_client.open_cluster 
      conf#nn_clustername
      (List.map Plasma_util.parse_host_port conf#nn_nodes)
      esys in
  let c = open_cluster (Unixqueue.create_unix_event_system()) in

  (* The idea of setting the max load for CPU to limit/2: This keeps
     capacity free for shuffle tasks when sorts are running. 
   *)

  let rjc =
    Mapred_job_exec.create_runtime_job_config
      ~io_load:conf#mr_task_load_limit
      ~cpu_load:(max 1.0 (conf#mr_task_load_limit /. 2.0))
      ~dump_plan_when_complete:!dump_plan
      ~shm_low:conf#mr_shm_low
      ~shm_high:conf#mr_shm_high
      mjc in
  (* TODO: allow to configure weight *)

  printf "Checking...\n%!";
  List.iter
    (fun dir ->
       try
	 Plasma_client.with_trans c
	   (fun trans ->
	      let files = Plasma_client.list trans dir in
	      if files <> [] then
		failwith ("Directory not empty: " ^ dir)
	   )
       with
	 | error ->
	     failwith ("PlasmaFS exception: " ^ Netexn.to_string error)
    )
    [ jc#output_dir; jc#work_dir; jc#log_dir ];

  let me =
    ( object
	method open_cluster = open_cluster
	method config = conf
	method config_file = cf
      end
    ) in
  mj # check_config me jc;

  printf "Planning...\n%!";

  let pc = 
    Mapred_sched.configure_plan
      ~keep_temp_files:!keep_temp_files 
      jc
      conf in

  let plan = Mapred_sched.create_plan c pc in
  Mapred_sched.add_inputs plan;

  if !dump_plan then (
    printf "INITIAL PLAN UP TO MAP:\n";
    Mapred_sched.print_plan plan
  );

  let tm = Mapred_taskfiles.scp_taskfile_manager conf jc in
  if jc#task_files <> [] then (
    printf "Copying task files...\n%!";
    tm # copy_to_local_directory jc#task_files
  );

  printf "Pre-start hook...\n%!";
  mj # pre_job_start me jc;

  printf "Starting...\n%!";
  let esys = Unixqueue.create_unix_event_system() in
  let job = Mapred_job_exec.start esys plan mj rjc conf#mr_task_port in
      
  let interrupt = ref false in
  Netsys_signal.register_handler
    ~name:"ctrl-c" ~signal:Sys.sigint 
    ~callback:(fun _ -> interrupt := true)
    ();

  printf "Job can be interrupted with CTRL-C!\n%!";
  
  let g = Unixqueue.new_group esys in
  let rec ctrlc() =
    Unixqueue.weak_once esys g 0.1
      (fun () ->
	 if !interrupt then (
	   printf "CTRL-C seen. Killing job.\n%!";
	   Mapred_job_exec.kill job
	 ) else
	   ctrlc()
      ) in
  ctrlc();
  
  Unixqueue.run esys;
  printf "Job_status: %s\n%!"
    (match Mapred_job_exec.status job with
       | `Running -> "running (hey, this is weird)"
       | `Successful -> "successful"
       | `Errors e -> "erroneous (" ^ String.concat "; " e ^ ")"
       | `Killed -> "killed"
    );

  printf "Cleanup...\n%!";
  Mapred_job_exec.cleanup job;
  if not !keep_temp_files then
    tm # delete_local_directory();  (* even if task_files = [] *)

  printf "Post-finish hook...\n%!";
  mj # post_job_finish me jc;

  printf "Finished.\n%!"


let dispatch mj =
  if Array.length Sys.argv < 2 then
    failwith "Not enough arguments";

  let mode_str = Sys.argv.(1) in
  incr Arg.current;

  match mode_str with
    | "start_task_servers" | "start-task-servers" ->
	start_task_servers mj
    | "stop_task_servers" | "stop-task-servers" ->
	stop_task_servers mj
    | "exec_job" | "exec-job" ->
	exec_job mj
    | "task_server" ->
	task_server mj
    | "help" | "-help" | "--help" -> usage()
    | _ -> failwith ("Bad mode")



let main mj =
  try
    Printexc.record_backtrace true;
    dispatch mj
  with
    | Failure m ->
        prerr_endline (this ^ ": " ^ m);
        usage();
        exit 2
          
