(*
  Copyright 2010 Gerd Stolpmann

  This file is part of Plasma, a distributed filesystem and a
  map/reduce computation framework. Unless you have a written license
  agreement with the copyright holder (Gerd Stolpmann), the following
  terms apply:

  Plasma is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Plasma is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.

*)
(* $Id: plasma_util.ml 273 2010-10-24 01:54:02Z gerd $ *)

(** Modules *)

module I64Set = Set.Make(Int64)

(** Utility functions *)

open Printf

let parse_host_port s =
  try
    Scanf.sscanf s "%s@:%d%!" 
      (fun h p -> (h,p))
  with
    | _ -> failwith "Util.parse_host_port"

let host_port_of_sockaddr =
  function
    | Unix.ADDR_INET(ip,p) ->
	(Unix.string_of_inet_addr ip, p)
    | Unix.ADDR_UNIX s ->
	failwith "Util.host_port_of_sockaddr"

let ip_of_host h =
  try Unix.inet_addr_of_string h
  with _ ->
    ( try
	let e = Unix.gethostbyname h in
	e.Unix.h_addr_list.(0)
      with
	| Not_found ->
	    failwith ("Host not found: " ^ h) 
    )

let sockaddr_of_host_port (h,p) =
  let ip = ip_of_host h in
  Unix.ADDR_INET(ip,p)


let connector_of_sockaddr =
  Rpc_client.connector_of_sockaddr


let ip_of_sockaddr addr =
  match addr with
    | Unix.ADDR_INET(ip,_) -> ip
    | Unix.ADDR_UNIX _ -> Unix.inet_addr_loopback (* FIXME: IPv6 not supp *)


type errno =
    (* everything but `ok *)
    [ `enotrans | `efailedcommit | `elongtrans | `efailed | `eperm
    | `enoent | `eaccess | `eexist | `efhier | `einval | `efbig
    | `enospc | `erofs | `enametoolong | `econflict | `ecoord
    | `enonode | `etbusy | `eio | `estale | `eloop | `enotdir 
    | `eisdir | `enotempty | `ebadpath
    ]

let string_of_errno te =
  match te with
    | `enotrans -> "ENOTRANS"
    | `efailedcommit -> "EFAILEDCOMMIT"
    | `elongtrans -> "ELONGTRANS"
    | `efailed -> "EFAILED"
    | `eperm -> "EPERM"
    | `enoent ->  "ENOENT"
    | `eaccess -> "EACCESS"
    | `eexist -> "EEXIST"
    | `efhier -> "EFHIER"
    | `einval -> "EINVAL"
    | `efbig -> "EFBIG"
    | `enospc -> "ENOSPC"
    | `erofs -> "EROFS"
    | `enametoolong -> "ENAMETOOLONG"
    | `econflict -> "ECONFLICT"
    | `ecoord -> "ECOORD"
    | `enonode -> "ENONODE"
    | `etbusy -> "ETBUSY" 
    | `eio -> "EIO"
    | `estale -> "ESTALE"
    | `eloop -> "ELOOP"
    | `enotdir -> "ENOTDIR"
    | `eisdir -> "EISDIR"
    | `enotempty -> "ENOTEMPTY"
    | `ebadpath -> "BADPATH"

exception Plasma_error of errno

let fallocate fd pos len =
  if len < 0L then
    invalid_arg "Plasma_util.fallocate";
  if len > Int64.sub Int64.max_int pos then 
    invalid_arg "Plasma_util.fallocate";
  try
    if not(Netsys_posix.have_fallocate()) then
      raise (Unix.Unix_error(Unix.ENOSYS,"",""));
    Netsys_posix.fallocate fd pos len
  with
    | Unix.Unix_error(Unix.ENOSYS,_,_) ->
	(* emulation. Note that this temporarily modifies the file position
	   (no pwrite)
	 *)
	let old_pos = Unix.LargeFile.lseek fd 0L Unix.SEEK_CUR in
	let eof =
	  Unix.LargeFile.lseek fd 0L Unix.SEEK_END in
	let s = "\000" in
	let k_start =
	  if pos < eof then Int64.sub eof pos else 0L in
	let k = ref k_start in
	while !k < len do
	  let p = Int64.add pos !k in
	  if p >= eof then (
	    ignore(Unix.LargeFile.lseek fd p Unix.SEEK_SET);
	    ignore(Unix.single_write fd s 0 1);
	  );
	  k := Int64.add !k 512L
	done;
	let new_eof = Int64.add pos len in
	if new_eof > eof then
	  Unix.LargeFile.ftruncate fd new_eof;
	ignore(Unix.LargeFile.lseek fd old_pos Unix.SEEK_SET)


(* read config file support : *)

let comment_re = Pcre.regexp "[^ \t]*#.*$"

let strip_comments s =
  Pcre.qreplace ~rex:comment_re ~templ:"" s


let node_list (cf:Netplex_types.config_file) addr =
  let port_opt =
    try
      let p = cf#resolve_parameter addr "port" in
      Some(cf#int_param p)
    with Not_found -> None in
  let parse_host_port s =
    try
      parse_host_port s
    with _ ->
      ( match port_opt with
	  | None -> 
	      failwith("Not in <host>:<port> syntax (no default port given): " 
			 ^ s)
	  | Some p ->
	      (s,p) 
      ) in
  let nodes =
    cf#resolve_section addr "node" in
  let direct_addrs =
    List.map
      (fun n ->
	 try
	   let p = cf#resolve_parameter n "addr" in
	   let a = cf#string_param p in
	   let (h,p) = parse_host_port a in
	   sprintf "%s:%d" h p
	 with
	   | Not_found ->
	       failwith ("missing 'addr' parameter in '" ^ cf#print n ^ "'")
      )
      nodes in
  let node_list_opt =
    try
      let p = cf#resolve_parameter addr "node_list" in
      Some(cf#string_param p)
    with Not_found -> None in
  match node_list_opt with
    | None ->
	if nodes = [] then
	  failwith ("missing 'node' or 'node_list' in '" ^ cf#print addr ^ "'");
	direct_addrs

    | Some node_list_file ->
	let fn =
	  if Filename.is_implicit node_list_file then
	    Filename.dirname cf#filename ^ "/" ^ node_list_file
	  else
	    node_list_file in
	let f = open_in fn in
	let addrs = ref [] in
	( try
	    while true do
	      let line = input_line f in
	      let line' = strip_comments line in
	      if line' <> "" then (
		let (h,p) = parse_host_port line' in
		addrs := (sprintf "%s:%d" h p) :: !addrs
	      )
	    done
	  with
	    | End_of_file -> ()
	);
	close_in f;

	direct_addrs @ List.rev !addrs





(**********************************************************************)
(* shared memory                                                      *)
(**********************************************************************)

exception Out_of_shared_memory

let rec get_shm_fd prefix n =
  let r = Digest.to_hex(Plasma_rng.random_bytes 16) in
  let name = Printf.sprintf "/%s_%s_%d" prefix r n in
  try
    (Netsys_posix.shm_open
       name [Netsys_posix.SHM_O_RDWR; 
	     Netsys_posix.SHM_O_CREAT;
	     Netsys_posix.SHM_O_EXCL ] 0o600,
     name)
  with
    | Unix.Unix_error(Unix.EEXIST,_,_) ->
	get_shm_fd prefix (n+1)


let open_shm_fd name =
  Netsys_posix.shm_open 
    name
    [ Netsys_posix.SHM_O_RDWR ]
    0


let enlarge_shm fd size =
  (* one must never shrink shm! Risk of sigbus! *)
  Unix.ftruncate fd size;
  try
    fallocate fd 0L (Int64.of_int size)
  with
    | Unix.Unix_error(Unix.ENOSPC,_,_) ->
	raise Out_of_shared_memory


let map_shm fd pos len =
  (* we try to avoid here an implicit enlargement of the shm file from
     map_file. This may turn out as race condition when several processes
     do the same with the same shm file, and suddenly one process finds
     them file truncated by the other...
   *)
  if pos < 0L || len < 0 then
    invalid_arg "Plasma_util.map_shm";
  let eof = Unix.LargeFile.lseek fd 0L Unix.SEEK_END in
  if Int64.sub eof (Int64.of_int len) < pos then
    failwith "Plasma_util.map_shm: shm file not large enough";
  Bigarray.Array1.map_file 
    fd ~pos Bigarray.char Bigarray.c_layout true len


(**********************************************************************)
(* Helpers for async code                                             *)
(**********************************************************************)

(* Old definitions now moved to ocamlnet: *)

let signal_engine  =
  Uq_engines.signal_engine

let fmap_engine = 
  Uq_engines.fmap_engine

let msync_engine l esys =
  Uq_engines.msync_engine l (fun _ () -> ()) () esys

let delay_engine =
  Uq_engines.delay_engine

let failsafe_e =
  Uq_engines.meta_engine


let string_of_state = 
  Uq_engines.string_of_state

let rpc_engine  =
  Rpc_proxy.ManagedClient.rpc_engine


let q_seq_engine queue esys f =
  Uq_engines.stream_seq_engine
    () 
    (Stream.from 
       (fun _ -> 
	  if Queue.is_empty queue then
	    None
	  else
	    let x = Queue.take queue in
	    Some(fun () -> f x)
       ))
    esys


let abort_if_working eng =
  match eng#state with
      `Working _ ->
	eng # abort()
    | _ ->
	()
;;


let trace_engine id e =
  let () =
    Netlog.logf `Debug
      "trace_engine %s: creating" id in
  ( object
      method request_notification f =
	Netlog.logf `Debug
	  "trace_engine %s: request_notification" id;
	e#request_notification 
	  (fun () ->
	     Netlog.logf `Debug
	       "trace_engine %s: notify (state %s)" 
	       id (string_of_state e#state);
	     try
	       let keep = f() in
	       Netlog.logf `Debug
		 "trace_engine %s: notify back (keep=%b)" id keep;
	       keep
	     with
	       | err ->
		   Netlog.logf `Debug
		     "trace_engine %s: notify exception %s"
		     id (Netexn.to_string err);
		   raise err
	  )
      method state = e#state
      method abort = e#abort
      method event_system = e#event_system
    end
  )

let ignore_lost_connection emit x =
  (** Call the [emit] function and ignore [Connection_lost] exceptions.
      This is required when [emit] is invoked after some delay, and 
      not immediately after receiving the request
   *)
  try emit x 
  with Rpc_server.Connection_lost -> ()


let rpc_srv_reply ?(onerror=fun _ -> ())
                  ?(onaborted=fun _ -> ())
                  ?(ondone=fun _ -> ())
                  procname sess emit 
                  (te : _ Uq_engines.final_state Uq_engines.engine) =
  (** [rpc_srv_reply name sess emit te]: Watches the engine [te] (as
      returned by [trans_e]), and when [te] finishes, the result is
      passed back as result of the RPC.

      The engine [te] must follow the conventions for transactional
      engines. Such engines never transition to [`Error] or [`Abort],
      but always to [`Done] (corresponding to the fact that transactions
      handle errors so they do not propagate automatically). The [`Done]
      state has the original state value as argument, i.e. we see
      [`Done(`Done v)], [`Done(`Error e)], and [`Done `Aborted], so
      the value of the inner engine enclosed by the transaction handler is
      still visible.

   *)
  Uq_engines.when_state
    ~is_done:(function
		| `Done arg -> 
		    ignore_lost_connection emit arg;
		    ondone arg
		| `Error error -> 
		    Netlog.logf `Err
		      "Error in %s: %s" procname (Netexn.to_string error);
		    ignore_lost_connection 
		      (Rpc_server.reply_error sess) Rpc.System_err;
		    onerror error
		| `Aborted ->
		    Netlog.logf `Err
		      "Execution of %s was aborted" procname;
		    ignore_lost_connection 
		      (Rpc_server.reply_error sess) Rpc.System_err;
		    onaborted()
	     )
    ~is_error:(fun error ->
		 Netlog.logf `Err
		   "Error in %s: %s" procname (Netexn.to_string error);
		 ignore_lost_connection 
		   (Rpc_server.reply_error sess) Rpc.System_err;
		 onerror error
	      )
    ~is_aborted:(fun () ->
		   Netlog.logf `Err
		     "Execution of %s was aborted" procname;
		   ignore_lost_connection 
		     (Rpc_server.reply_error sess) Rpc.System_err;
		   onaborted()
		)
    te


module Operators = struct
  let map_e = new Uq_engines.map_engine
  let eps_e = new Uq_engines.epsilon_engine
  let msync_e = msync_engine
  let qseq_e = q_seq_engine
  let ( ++ ) = new Uq_engines.seq_engine
  let ( >> ) = fmap_engine
    
(** Remember:
    
    - [ e1 ++ (fun r -> e2) ]
      means that we run [e1] first, and if it is successful with result [r],
      we start [e2]
    - [ e1 >> (fun s1 -> s2) ]
      means that we run [e1], and any result state [s1] is mapped into [s2]
      (including error states)
    - [ eps_e x esys ]
      "lifts" the value [x] into engine calculus
    - [ msync_e [e1;e2;...] esys ]
      waits until all argument engines e1,e2,... are finished
    - [ qseq_e q esys ]
      runs the engines in the queue [q] sequentially (destroying the queue)
 *)

end


(** Compute/verify reservation IDs *)

let extract64 s =
  (** Extract 64 bits (as [int64]) from a string with at least 8 bytes *)
  Rtypes.int64_of_int8(Rtypes.read_int8 s 0)


let compute_verifier res_id res_secret block =
  let s = 
    sprintf "%Ld/%Ld/%Ld" 
      res_id
      res_secret
      block in
  let h = Digest.string s in
  extract64 h


(** LRU cache *)

module LRU(Ord : Map.OrderedType) : sig
  type elt = Ord.t
  type t

  val create : unit -> t
    (** new cache *)

  val length : t -> int
    (** number of elements *)

  val is_empty : t -> bool
    (** whether empty *)

  val add : elt -> t -> unit
    (** adds a new element, or if it is already in the cache, marks it as
       "just used"
     *)

  val take : t -> elt
    (** takes the least recently used element (or raises Not_found) *)

  val mem : elt -> t -> bool
    (** whether member *)

  val clear : t -> unit
    (** removes all elements *)

  val remove : elt -> t -> unit
    (** removes an arbitrary element *)

end = struct
  type elt = Ord.t

  module M = Map.Make(Ord)
  module IS = Set.Make(Int64)
  module IM = Map.Make(Int64)

  type t =
      { mutable length : int;
	mutable max_age : int64;
	mutable age_of_elt : int64 M.t;
	mutable elt_of_age : elt IM.t;
	mutable ages : IS.t;
      }
   (* Note negative logic: if the age value grows, we consider the
      entry to be younger
    *)

  let create() =
    { length = 0;
      max_age = 0L;
      age_of_elt = M.empty;
      elt_of_age = IM.empty;
      ages = IS.empty;
    }

  let clear lru =
    lru.length <- 0;
    lru.max_age <- 0L;
    lru.age_of_elt <- M.empty;
    lru.elt_of_age <- IM.empty;
    lru.ages <- IS.empty

  let length lru =
    lru.length

  let is_empty lru =
    lru.length = 0

  let add p lru =
    ( try
	let age = M.find p lru.age_of_elt in
	lru.ages <- IS.remove age lru.ages;
	lru.elt_of_age <- IM.remove age lru.elt_of_age;
	lru.length <- lru.length - 1
      with
	| Not_found -> ()
    );
    let new_age = Int64.succ lru.max_age in
    lru.max_age <- new_age;
    lru.ages <- IS.add new_age lru.ages;
    lru.age_of_elt <- M.add p new_age lru.age_of_elt;
    lru.elt_of_age <- IM.add new_age p lru.elt_of_age;
    lru.length <- lru.length + 1
    
  let remove p lru =
    try
      let age = M.find p lru.age_of_elt in
      lru.ages <- IS.remove age lru.ages;
      lru.elt_of_age <- IM.remove age lru.elt_of_age;
      lru.age_of_elt <- M.remove p lru.age_of_elt;
      lru.length <- lru.length - 1
    with
      | Not_found -> ()

  let take lru =
    let age = IS.min_elt lru.ages in
    let p = IM.find age lru.elt_of_age in
    remove p lru;
    p

  let mem p lru =
    M.mem p lru.age_of_elt

end


(** Find ranges *)

(** Problem here is that there is a set of integers, and these should not
    be handled as individual elements of a set, but ranges of ints are
    the elements
 *)

module type ExtOrderedType = sig
  type t
  val compare : t -> t -> int
  val pred : t -> t option
  val succ : t -> t option
end

module RangeSet(Ord : ExtOrderedType) : sig 
  type elt = Ord.t

  type t

  val empty : t

  val add : elt -> t -> t
    (** [add x s]: Returns the set where [x] is a new element *)
	
  val iter : (elt -> elt -> unit) -> t -> unit
    (** [iter f s]: applies [f] to all ranges in [s], i.e. pairs
	[(r1,r2)] so that all elements from [r1] to [r2] are member
	of the set, and [pred r1] is not a member, and [succ r2]
	is not a member either.
     *)
end = struct

  module M = Map.Make(Ord)

  type elt = Ord.t

  type t =
      { max_of_min : elt M.t;
	min_of_max : elt M.t;
	member : unit M.t
      }

  let empty =
    { max_of_min = M.empty;
      min_of_max = M.empty;
      member = M.empty
    }

  let add x s =
    if M.mem x s.member then
      s
    else (
      let preceding_range =
	match Ord.pred x with
	  | None -> None
	  | Some px ->
	      if M.mem px s.member then
		Some (M.find px s.min_of_max, px)
	      else
		None in
      let succeeding_range =
	match Ord.succ x with
	  | None -> None
	  | Some sx ->
	      if M.mem sx s.member then
		Some (sx, M.find sx s.max_of_min)
	      else
		None in
      match preceding_range, succeeding_range with
	| None, None ->
	    { max_of_min = M.add x x s.max_of_min;
	      min_of_max = M.add x x s.min_of_max;
	      member = M.add x () s.member
	    }
	| (Some(pmin,pmax)), None ->
	    { max_of_min = M.add pmin x s.max_of_min;
	      min_of_max = M.add x pmin (M.remove pmax s.min_of_max);
	      member = M.add x () s.member
	    }
	| None, (Some(smin,smax)) ->
	    { max_of_min = M.add x smax (M.remove smin s.max_of_min);
	      min_of_max = M.add smax x s.min_of_max;
	      member = M.add x () s.member
	    }
	| (Some(pmin,pmax)), (Some(smin,smax)) ->
	    { max_of_min = M.add pmin smax (M.remove smin s.max_of_min);
	      min_of_max = M.add smax pmin (M.remove pmax s.min_of_max);
	      member = M.add x () s.member
	    }
    )

  let iter f s =
    M.iter f s.max_of_min
  
end
